#include "pdp2.h"
#include "xgrafix.h"

/*----------------------------------------------------------------------------

This mcc package requires the input file to specify species with the names:

U238_100 = U238 neutrals (rel. weight = 100)
electrons (rel. weight = 1)
U235P = U235 ions (rel . weight = 1)
U235 = U235 neutrals (rel. weight = 1)
U238P = U238 ions (rel. weight = 1)
U238 = U238 neutrals (rel. weight = 1) 

These names can have other characters appended to the end, in order to make
additonal species of the same type (e.g. 238_inj and U238_bg). Note that mcc
will only find the first species for each initial string above (e.g. U238_inj for
the previous example, such that any U238P which charge exchange will create 
U238_inj).

----------------------------------------------------------------------------*/

#define   NEMAX   500
#define   NIMAX   500

/* following are used to determine the type of collisions for a species */
#define COL_NONE  0
#define COL_U238  1
#define COL_U238P 2
#define COL_U235  3
#define COL_U235P 4

void newvel(SCALAR energy, SCALAR vel, SCALAR *vx, SCALAR *vy, SCALAR *vz, int e_flag);
void scattering(int isp, int j, SCALAR vel);

int ecolsp;
SCALAR gden, vgth, extengy0, ionengy0;
SCALAR sigma_el(SCALAR energy), sigma_cx(SCALAR energy);

/* SCALAR sigma1(SCALAR energy), sigma2(SCALAR energy), sigma3(SCALAR energy);
	SCALAR sigma4(SCALAR energy), sigma5(SCALAR energy);
	int ecolsp, icolsp;
	*/

/* these are the species indices for the respective ions and neutrals: */
int U235P, U238P, U235, U238, U238_100;

/*********************************************************/
/* Monte Carlo Collisions for electron- and ion-neutrals */

void mcc(int isp)
{
  static int init_mcc_flag=1;
  static SCALAR *icol_extra;
  static SCALAR *max_sigmav;
  static SCALAR *col_prob;
  static int *collisionType;

  register int i, j, k, index;
  int N, nnp, s1, s2;
  SCALAR random, vel, sigma_total;
  SCALAR engy, rengy, delx, dely;
  SCALAR sum_sigma, temp, vneutx, vneuty, vneutz;

  if (init_mcc_flag) {
#ifdef OLD_CODE
    gden = NperTORR*pressure/(gtemp+SNG_MIN);  /* Calculating the gas density */
    
	 ecolsp = ecollisional-1;
    ionsp--;
    if(ionsp>0) vgth= sqrt(2*1.602e-19*gtemp/m[ionsp]);
#endif
	 icol_extra = (SCALAR*) malloc(nsp*sizeof(SCALAR));
	 max_sigmav = (SCALAR*) malloc(nsp*sizeof(SCALAR));
	 col_prob = (SCALAR*) malloc(nsp*sizeof(SCALAR));
	 collisionType = (int*) malloc(nsp*sizeof(int));
	 
	 U235P = U238P = U235 = U238 = U238_100 = -1; /* init with invalid value */
	 /* loop in reverse order so the variabkes are set to the first match */
	 for (i=nsp-1; i>=0; i--) {
		if (strncmp(name[i], "U235P", 5)==0) { /* check longer name first */
		  U235P = isp;
		  collisionType[i] = COL_U235P;
		}
		else if (strncmp(name[i], "U235", 4)==0) {
		  U235 = i;
		  collisionType[i] = COL_U235;
		} 
		else if (strncmp(name[i], "U238_100", 8)==0) {
		  U238_100 = i;
		  collisionType[i] = COL_U238;
		}
		else if (strncmp(name[i], "U238P", 5)==0) {
		  U238P = i;
		  collisionType[i] = COL_U238P;
		}
		else if (strncmp(name[i], "U238", 4)==0) {
		  U238 = i;
		  collisionType[i] = COL_U238;
		}
		else collisionType[i] = COL_NONE;
	 }

	 if (U238_100 < 0) {
		printf("Deactivating collisions - you need a U238_100 species for the uranium package\n");
		icollisional = 0;
		return;
	 }
    
    /*****************************************/
    /* Calculating the null collision prob.  */

	 /* currently NO electron collisions in this package */
    
    if (icollisional) {
		for (i=0; i < nsp; i++) {
		  SCALAR ftemp;
		  max_sigmav[i] = 0;
		  for (engy=0; engy < NIMAX; engy += 0.1) {
			 switch(collisionType[i]) {
			 case COL_U235P: ftemp = sigma_el(engy) + sigma_cx(engy);
				break;
			 case COL_U235: ftemp = sigma_el(engy);
				break;
			 case COL_U238: ftemp = sigma_el(engy);
				break;
			 case COL_U238P: ftemp = sigma_el(engy) + sigma_cx(engy);
				break;
			 }
			 max_sigmav[i] = max(max_sigmav[i], sqrt(engy)*ftemp);
		  }
		  max_sigmav[i] *= sqrt(2*1.602E-19/m[i]);
		  icol_extra[i] = 0.5;
		}
		init_mcc_flag = 0;
	 }
  }
  
  if (icollisional) {
	 SCALAR max_n238;
	 max_n238 = 0;
	 for (i=0; i < ngx; i++) 
		for (j=0; j < ngy; j++) 
		  max_n238 = max(max_n238, sp_n[U238_100][i][j]);
	 max_n238 *= nc2p/(zlength*dx*dy); /* normalize to MKS units */
	 col_prob[isp] = 1 - exp(-max_sigmav[isp]*max_n238*dt*sp_k[isp]);
	 icol_extra[isp] += np[isp]*col_prob[isp];
	 N = icol_extra[isp]; /* this is the number of species isp to collide this step */
	 icol_extra[isp] -= N;
	 nnp = np[isp];
	 for (i=0; i < N; i++) {
		index = nnp*frand();
		nnp--;

		temp = x[isp][nnp];
		x[isp][nnp] = x[isp][index];
		x[isp][index] = temp;

		temp = y[isp][nnp];
		y[isp][nnp] = y[isp][index];
		y[isp][index] = temp;

		temp = vx[isp][nnp];
		vx[isp][nnp] = vx[isp][index];
		vx[isp][index] = temp;

		temp = vy[isp][nnp];
		vy[isp][nnp] = vy[isp][index];
		vy[isp][index] = temp;

		temp = vz[isp][nnp];
		vz[isp][nnp] = vz[isp][index];
		vz[isp][index] = temp;
	 }
	 
	 for (j=nnp; j < nnp+N; j++) {
		SCALAR vth;
		SCALAR wx, wy; /* weight in x and y for current particle */
		int jx, jy; /* cell indices for current particle */
		jx = x[isp][j];
		wx = x[isp][j] - jx;
		jy = y[isp][j];
		wy = y[isp][j] - jy;
		vth = (sp_vt[U238_100][jx][jy]*(1-wx) + sp_vt[U238_100][jx+1][jy]*wx)*(1-wy)
		  + (sp_vt[U238_100][jx][jy+1]*(1-wx) + sp_vt[U238_100][jx+1][jy+1]*wx)*wy;
      maxwellv(&vneutx, &vneuty, &vneutz, vth, frand(), frand(), frand());
		vneutx += (sp_vx0[U238_100][jx][jy]*(1-wx) + sp_vx0[U238_100][jx+1][jy]*wx)*(1-wy)
		  + (sp_vx0[U238_100][jx][jy+1]*(1-wx) + sp_vx0[U238_100][jx+1][jy+1]*wx)*wy;
		vneuty += (sp_vy0[U238_100][jx][jy]*(1-wx) + sp_vy0[U238_100][jx+1][jy]*wx)*(1-wy)
		  + (sp_vy0[U238_100][jx][jy+1]*(1-wx) + sp_vy0[U238_100][jx+1][jy+1]*wx)*wy;
		vneutz += (sp_vz0[U238_100][jx][jy]*(1-wx) + sp_vz0[U238_100][jx+1][jy]*wx)*(1-wy)
		  + (sp_vz0[U238_100][jx][jy+1]*(1-wx) + sp_vz0[U238_100][jx+1][jy+1]*wx)*wy;
      vneutx /= vxscale[U238_100];
      vneuty /= vyscale[U238_100];
      vneutz /= vzscale[U238_100];
      vx[isp][j] -= vneutx; /* frame of neutral */
      vy[isp][j] -= vneuty;
      vz[isp][j] -= vneutz;
      engy = Exscale[isp]*vx[isp][j]*vx[isp][j];
      engy+= Eyscale[isp]*vy[isp][j]*vy[isp][j];
      engy+= Ezscale[isp]*vz[isp][j]*vz[isp][j];
      vel  = sqrt(2.0*1.602e-19*engy/m[isp]);
      sigma_total = max_sigmav[isp]/(vel + SNG_MIN);
      random= frand();

		if (collisionType[isp] == COL_U235P) {
		  if (random <= (sum_sigma = sigma_el(engy))/sigma_total) {
			 scattering(isp, j, vel);
			 vx[isp][j] += vneutx; /* back to the lab frame */
			 vy[isp][j] += vneuty;
			 vz[isp][j] += vneutz;
		  }
		  else if (random <= (sum_sigma += sigma_cx(engy))/sigma_total) {
			 /* first, create fast neutral U235 */
			 vx[U235][np[U235]] = vx[isp][j] + vneutx; 
			 vy[U235][np[U235]] = vy[isp][j] + vneuty; 
			 vz[U235][np[U235]] = vz[isp][j] + vneutz;
			 x[U235][np[U235]] = x[isp][j];
			 y[U235][np[U235]] = y[isp][j];
			 np[U235]++;
			 /* next, create the slow U238P */
			 vx[U238P][np[U238P]] = vneutx; 
			 vy[U238P][np[U238P]] = vneuty; 
			 vz[U238P][np[U238P]] = vneutz;
			 x[U238P][np[U238P]] = x[isp][j];
			 y[U238P][np[U238P]] = y[isp][j];
			 np[U238P]++;
			 /* finally, move the last U235P into this slot and retest */
			 vx[isp][j] = vx[isp][nnp+N-1];
			 vy[isp][j] = vy[isp][nnp+N-1];
			 vz[isp][j] = vz[isp][nnp+N-1];
			 x[isp][j] = x[isp][nnp+N-1];
			 y[isp][j] = y[isp][nnp+N-1];
			 np[isp]--;
			 N--;
			 j--; /* roll back j; it will get ++ in the for loop to retest this particle */
		  }
		}
		else if (collisionType[isp] == COL_U235) {
		  if (random <= (sum_sigma = sigma_el(engy))/sigma_total) {
			 scattering(isp, j, vel);
			 vx[isp][j] += vneutx; /* back to the lab frame */
			 vy[isp][j] += vneuty;
			 vz[isp][j] += vneutz;
		  }
		}
		else if (collisionType[isp] == COL_U238P) {
		  if (random <= (sum_sigma = sigma_el(engy))/sigma_total) {
			 scattering(isp, j, vel);
			 vx[isp][j] += vneutx; /* back to the lab frame */
			 vy[isp][j] += vneuty;
			 vz[isp][j] += vneutz;
		  }
		  else if (random <= (sum_sigma += sigma_cx(engy))/sigma_total) {
			 /* first, create the fast neutral U238 */
			 vx[U238][np[U238]] = vx[isp][j] + vneutx; 
			 vy[U238][np[U238]] = vy[isp][j] + vneuty; 
			 vz[U238][np[U238]] = vz[isp][j] + vneutz;
			 x[U238][np[U238]] = x[isp][j];
			 y[U238][np[U238]] = y[isp][j];
			 np[U238]++;
			 /* next, create the slow U238P */
			 vx[isp][j] = vneutx;
			 vy[isp][j] = vneuty;
			 vz[isp][j] = vneutz;
		  }
		}
		else if (collisionType[isp] == COL_U238) {
		  if (random <= (sum_sigma = sigma_el(engy))/sigma_total) {
			 scattering(isp, j, vel);
			 vx[isp][j] += vneutx; /* back to the lab frame */
			 vy[isp][j] += vneuty;
			 vz[isp][j] += vneutz;
		  }
		}
    }
  }
}

/**************************************************************/

/**********************************/
/* if the collision is scattering */

void scattering(int isp, int j, SCALAR vel)
{      
  SCALAR up1, up2, up3, mag;
  SCALAR r11, r12, r13, r21, r22, r23, r31, r32, r33;
  SCALAR cosphi, sinphi, coschi, sinchi, phi1;
  
  coschi= sqrt(frand());
  sinchi= sqrt(fabs(1. -coschi*coschi));
		
  phi1  = 2*M_PI*frand();
  cosphi= cos(phi1);
  sinphi= sin(phi1);
		
  r13 = vx[isp][j]*vxscale[isp]/vel;
  r23 = vy[isp][j]*vyscale[isp]/vel;
  r33 = vz[isp][j]*vzscale[isp]/vel;
	
  if(r33 == 1.0) { up1= 0;  up2= 1;  up3= 0; }
  else           { up1= 0;  up2= 0;  up3= 1; }
  
  r12 = r23*up3 -r33*up2;
  r22 = r33*up1 -r13*up3;
  r32 = r13*up2 -r23*up1;
  mag = sqrt(r12*r12 + r22*r22 + r32*r32);
	
  r12/= mag;
  r22/= mag;
  r32/= mag;
	
  r11 = r22*r33 -r32*r23;
  r21 = r32*r13 -r12*r33;
  r31 = r12*r23 -r22*r13;
	
  vx[isp][j]= vel*coschi*(r11*sinchi*cosphi +r12*sinchi*sinphi +r13*coschi)/vxscale[isp];
  vy[isp][j]= vel*coschi*(r21*sinchi*cosphi +r22*sinchi*sinphi +r23*coschi)/vyscale[isp]; 
  vz[isp][j]= vel*coschi*(r31*sinchi*cosphi +r32*sinchi*sinphi +r33*coschi)/vzscale[isp];
}

void newvel(SCALAR energy, SCALAR vel, SCALAR *vx, SCALAR *vy, SCALAR *vz, int e_flag)
{
  SCALAR phi1, cosphi, sinphi, coschi, sinchi, up1, up2, up3;
  SCALAR mag, r11, r12, r13, r21, r22, r23, r31, r32, r33;

  if(energy < 1e-30)  coschi = 1;
  else  coschi = (energy +2 -2*pow(energy +1,frand()))/energy;
  sinchi= sqrt(fabs(1. - coschi*coschi));
  
  phi1  = 2*M_PI*frand();
  cosphi= cos(phi1);
  sinphi= sin(phi1);
  
  if(e_flag) vel *= sqrt(1.0 - 2.0*m[ecolsp]*(1-coschi)/m[ionsp]);
  
  r13 = *vx;
  r23 = *vy;
  r33 = *vz;
  
  if(r33 == 1.0) { up1= 0;  up2= 1;  up3= 0; }
  else           { up1= 0;  up2= 0;  up3= 1; }
  
  r12 = r23*up3 -r33*up2;
  r22 = r33*up1 -r13*up3;
  r32 = r13*up2 -r23*up1;
  mag = sqrt(r12*r12 + r22*r22 + r32*r32);
  
  r12/= mag;
  r22/= mag;
  r32/= mag;
  
  r11 = r22*r33 -r32*r23;
  r21 = r32*r13 -r12*r33;
  r31 = r12*r23 -r22*r13;
  
  *vx= vel*(r11*sinchi*cosphi +r12*sinchi*sinphi +r13*coschi);
  *vy= vel*(r21*sinchi*cosphi +r22*sinchi*sinphi +r23*coschi); 
  *vz= vel*(r31*sinchi*cosphi +r32*sinchi*sinphi +r33*coschi);
}

/* U elastic scattering */

SCALAR sigma_el(SCALAR E)
{
  return 2.4E-19;
}

/* U charge exchange */

SCALAR sigma_cx(SCALAR E)
{
  return 1E-20*pow(12.6 - 1.55*log10(E),2) + 1.153E-20*pow(E,-1.12);
}

/**************************************************************/
